/*    $Id: ExprNode.java,v 1.1 2002/09/02 05:39:34 gage Exp $
**
**
**************************************************************************
**    Copyright (C) 1996 David Wanqian Liu
**                       davidliu@mti.sgi.com
**                       http://reality.sgi.com/exmployees/davidliu_mti
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**************************************************************************
*/


package expression;



/**
  A reduced expression node in an operator-precedence parsing tree.
  */
public class ExprNode {
  /**
    An empty expression node.
    */
  final static int EXPNODE_EMPTY = 0;

  /**
    An expression node containing a single expression node.
    */
  final static int EXPNODE_NODE = 1;

  /**
    An expression node containing an expression node and
    a unary function applicable on it.
    */
  final static int EXPNODE_UNARYFUNC = 2;

  /**
    An expression node containing two expression nodes and
    a binary function applicable on them.
    */
  final static int EXPNODE_BINFUNC = 3;

  /**
    An expression node containing a constant number.
    */
  final static int EXPNODE_CONST = 4;

  /**
    An expression node containing a variable.
    */
  final static int EXPNODE_VAR = 5;

  /**
    Expression node type.
    */
  public int etype;

  /* Number of nodes contained in the expression node */
  int numNodes;

  /**
    Expression node 1 contained in the current expression node.
    It is valid when the type of current expression node is one of
    the following: EXPNODE_NODE, EXPNODE_UNARYFUNC and EXPNODE_BINFUNC.
    */
  ExprNode node1;

  /**
    Expression node 2 contained in the current expression node.
    It is valid when the type of current expression node is EXPNODE_BINFUNC.
    */
  ExprNode node2;

  /**
    Function contained in the current expression node.
    It is valid when the type of current expression node is either
    one of the following: EXPNODE_UNARYFUNC and EXPNODE_BINFUNC.
    */
  Function func;

  /**
    Constant contained in the current expression node.
    It is valid when the type of current expression node isEXPNODE_CONST.
    */
  double id;

  /**
    Variable contained in the current expression node.
    It is valid when the type of current expression node isEXPNODE_VAR.
    */
  Variable variable;


  /**
    Create a default empty expression node.
    */
  public ExprNode() {
    etype = EXPNODE_EMPTY;
    numNodes = 0;
    node1 = null;
    node2 = null;
    id = 0.0;
    variable = null;
  }


  /**
    Pushes a symbol into the expression node. It will dispatch
    corresponding operation depending on the type of input symbol.
    */
  public void push(Symbol sym) throws ExpressionParseException {
    switch (sym.stype) {
    default:
    case Symbol.SYM_UNKNOWN:
      throw new ExpressionParseException(ExpressionParseException.UNKNOWN_SYMBOL);

    case Symbol.SYM_EXPRNODE:
      push(sym.enode);
      return;

    case Symbol.SYM_TERMINAL:
      switch(sym.ttype) {
      default:
	throw new ExpressionParseException(ExpressionParseException.UNKNOWN_TERMINAL);

      case Symbol.TK_EOE:
	throw new ExpressionParseException(ExpressionParseException.UNEXPECTED_SYMBOL);

      case Symbol.TK_NUMBER:
	push(sym.nval);
	return;
	
      case Symbol.TK_VARIABLE:
	push(sym.variable);
	return;

      case Symbol.TK_BRACKETL:
      case Symbol.TK_BRACKETR:
	return;

      case Symbol.TK_PLUS:
      case Symbol.TK_MINUS:
      case Symbol.TK_MULT:
      case Symbol.TK_DIV:
      case Symbol.TK_POW:
      case Symbol.TK_FUNCTION:
	push(sym.func);
      }
    }
  }


  /**
    Pushes an expression node into the expression node.
    */
  public void push(ExprNode n) throws ExpressionParseException {
    switch (etype) {
    case EXPNODE_EMPTY:
      etype = EXPNODE_NODE;
      node1 = n;
      return;

    case EXPNODE_NODE:
      throw new ExpressionParseException(ExpressionParseException.NO_OPERATOR);

    case EXPNODE_UNARYFUNC:
      etype = EXPNODE_BINFUNC;
      node2 = n;
      return;

    case EXPNODE_BINFUNC:
      throw new ExpressionParseException(ExpressionParseException.TOOMANY_NODE);

    case EXPNODE_CONST:
    case EXPNODE_VAR:
      throw new ExpressionParseException(ExpressionParseException.UNEXPECTED_NODE);
      
    default:
      throw new ExpressionParseException();
    }
  }


  /**
    Pushes a function into the expression node.
    */
  public void push(Function f) throws ExpressionParseException {
    switch (etype) {
    default:
      throw new ExpressionParseException(ExpressionParseException.NO_NODE);

    case EXPNODE_NODE:
      etype = EXPNODE_UNARYFUNC;
      func = f;
      return;
    }
  }


  /**
    Pushes a const into the expression node.
    */
  public void push(double i) throws ExpressionParseException {
    switch (etype) {
    default:
      throw new ExpressionParseException(ExpressionParseException.UNEXPECTED_SYMBOL);

    case EXPNODE_EMPTY:
      etype = EXPNODE_CONST;
      id = i;
    }
  }


  /**
    Pushes a variable into the expression node.
    */
  public void push(Variable v) throws ExpressionParseException {
    switch (etype) {
    default:
      throw new ExpressionParseException(ExpressionParseException.UNEXPECTED_SYMBOL);

    case EXPNODE_EMPTY:
      etype = EXPNODE_VAR;
      variable = v;
    }
  }


  /**
    Evaluate the expression node.
    */
  public double eval() throws ExpressionEvalException {
    switch (etype) {
    case EXPNODE_EMPTY:
      throw new ExpressionEvalException(ExpressionEvalException.EMPTY_NODE);

    case EXPNODE_NODE:
      return node1.eval();

    case EXPNODE_UNARYFUNC:
      if (func.numArgs != 1) {
	if ("+".equals(func.toString()) || "-".equals(func.toString()))
	  return func.apply(0.0, node1.eval());
	else
	  throw new ExpressionEvalException(ExpressionEvalException.EXPECT_UNARYFUNC);
      }

      return func.apply(node1.eval());

    case EXPNODE_BINFUNC:
      if (func.numArgs != 2)
	throw new ExpressionEvalException(ExpressionEvalException.EXPECT_BINFUNC);
      
      return func.apply(node2.eval(), node1.eval());

    case EXPNODE_CONST:
      return id;

    case EXPNODE_VAR:
      return variable.val;
      
    default:
      throw new ExpressionEvalException();
    }
  }


  /**
    Returns the string representation of the expression node.
    */
  public String toString() {
    switch(etype) {
    case EXPNODE_EMPTY:
      return "Empty.";

    case EXPNODE_NODE:
      return "Single node.";

    case EXPNODE_UNARYFUNC:
      return "Unary function.";

    case EXPNODE_BINFUNC:
      return "Binary function.";

    case EXPNODE_CONST:
      return ExprDouble.toString(id);

    case EXPNODE_VAR:
      return variable.toString();
      
    default:
      return "Unknown.";
    }
  }
  
}
