/*    $Id: Symbol.java,v 1.1 2002/09/02 05:39:36 gage Exp $
**
**
**************************************************************************
**    Copyright (C) 1996 David Wanqian Liu
**                       davidliu@mti.sgi.com
**                       http://reality.sgi.com/exmployees/davidliu_mti
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program; if not, write to the Free Software
**    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**************************************************************************
*/


package expression;



/**
  An input symbol in an expression.
  */
public class Symbol {
  /**
    The symbol type representing unknown symbol.
    */
  public final static int SYM_UNKNOWN = 0;

  /**
    The symbol type representing a reduced expression node.
    */
  public final static int SYM_EXPRNODE = 1;

  /**
    The symbol type representing a terminal symbol.
    */
  public final static int SYM_TERMINAL = 2;

  /**
    The token representing the end of an expression.
    */
  public final static int TK_EOE = 1000;

  /**
    The token representing a number in an expression.
    */
  public final static int TK_NUMBER = 1001;

  /**
    The token representing a variable in an expression.
    */
  public final static int TK_VARIABLE = 1002;

  /**
    The token representing a left bracket in an expression.
    */
  public final static int TK_BRACKETL = 1003;

  /**
    The token representing a right bracket in an expression.
    */
  public final static int TK_BRACKETR = 1004;

  /**
    The token representing a plus sign in an expression.
    */
  public final static int TK_PLUS = 1005;

  /**
    The token representing a minus sign in an expression.
    */
  public final static int TK_MINUS = 1006;

  /**
    The token representing a multply sign in an expression.
    */
  public final static int TK_MULT = 1007;

  /**
    The token representing a division sign in an expression.
    */
  public final static int TK_DIV = 1008;

  /**
    The token representing a power sign in an expression.
    */
  public final static int TK_POW = 1009;

  /**
    The token representing a function in an expression.
    */
  public final static int TK_FUNCTION = 1010;

  /**
    Symbol type of the symbol.
    */
  public int stype;

  /**
    The terminal type of the symbol, if symbol type if SYM_TERMINAL.
    */
  public int ttype;

  /**
    The reduced expression node, if symbol type if SYM_EXPRNODE.
    */
  public ExprNode enode;

  /**
    The value of a TK_NUMBER token.
    */
  public double nval;

  /**
    The function of a TK_FUNCTION token.
    */
  public Function func;

  /**
    The variable of a TK_VARIABLE token.
    */
  public Variable variable;


  /**
    Precedence table used for parsing the expression.
    */
  static int precTab[][] = {
    /*  EOE   id   (    )    +    -    *    /    ^    f */
    {   'X', '<', '<', 'X', '<', '<', '<', '<', '<', '<'}, /*EOE */
    {   '>', 'X', 'X', '>', '>', '>', '>', '>', '>', 'X'}, /* id */
    {   'X', '<', '<', '=', '<', '<', '<', '<', '<', '<'}, /* ( */
    {   '>', 'X', 'X', '>', '>', '>', '>', '>', '>', 'X'}, /* ) */
    {   '>', '<', '<', '>', '>', '>', '<', '<', '<', '<'}, /* + */
    {   '>', '<', '<', '>', '>', '>', '<', '<', '<', '<'}, /* - */
    {   '>', '<', '<', '>', '>', '>', '>', '>', '<', '<'}, /* * */
    {   '>', '<', '<', '>', '>', '>', '>', '>', '<', '<'}, /* / */
    {   '>', '<', '<', '>', '>', '>', '>', '>', '<', '<'}, /* ^ */
    {   '>', '<', '<', '>', '>', '>', '>', '>', '>', '<'}, /* f */
  };

  /**
    Convert a token type to an index in the precedence table.
    */
  static int precTabIndex(int token)  throws ExpressionParseException {
    switch (token) {
    default:
      throw new ExpressionParseException("Unknown token");

    case TK_EOE:
      return 0;

    case TK_NUMBER:
    case TK_VARIABLE:
      return 1;

    case TK_BRACKETL:
      return 2;

    case TK_BRACKETR:
      return 3;

    case TK_PLUS:
      return 4;

    case TK_MINUS:
      return 5;

    case TK_MULT:
      return 6;

    case TK_DIV:
      return 7;

    case TK_POW:
      return 8;
      
    case TK_FUNCTION:
      return 9;
    }
  }


  /**
    Create an empty symbol.
    */
  public Symbol() {
    stype = SYM_UNKNOWN;
  }


  /**
    Create a token symbol with symbol type and token type.
    It is used for creating symbols for operators, EOE and brackets.
    */
  public Symbol(int type, int tk) {
    stype = type;
    ttype = tk;

    switch (tk) {
    default:
      return;

    case TK_PLUS:
      func = FunctionSet.getFunc("+");
      return;

    case TK_MINUS:
      func = FunctionSet.getFunc("-");
      return;

    case TK_MULT:
      func = FunctionSet.getFunc("*");
      return;

    case TK_DIV:
      func = FunctionSet.getFunc("/");
      return;

    case TK_POW:
      func = FunctionSet.getFunc("^");
      return;
    }
  }


  /**
    Create a Number token with symbol type specified as SYM_TERMINAL,
    token type specified as TK_NUMBER and a value
    */
  public Symbol(int type, int tk, double val) {
    stype = type;
    ttype = tk;
    nval = val;
  }


  public Symbol(int type, int tk, char vchar) throws ExpressionParseException {
    stype = type;
    ttype = tk;

  }



  /**
    Create a Variable token or Function token with symbol type specified
    as SYM_TERMINAL.
    Variable token will be created if token type is specified as
    TK_VARIABLE and the string representation of a variable ("X", "Y",
    "Z", "E", "PI") is given.
    Function token will be created if token type is specified as
    TK_FUNCTION and the string representation of the function is given.
    */
  public Symbol(int type, int tk, String str) throws ExpressionParseException {
    stype = type;
    ttype = tk;

    switch (ttype) {
    case TK_VARIABLE:
      if ("X".equals(str.toUpperCase()))
	variable = Variable.X;

      else if ("Y".equals(str.toUpperCase()))
	variable = Variable.Y;

      else if ("Z".equals(str.toUpperCase()))
	variable = Variable.Z;

      else if ("E".equals(str.toUpperCase()))
	variable = Variable.E;

      else if ("PI".equals(str.toUpperCase()))
	variable = Variable.PI;

      else
	throw new ExpressionParseException("Unrecognized Variable: " + str);

      return;

    case TK_FUNCTION:
      func = FunctionSet.getFunc(str);

      if (func == null)
	throw new ExpressionParseException("Unrecognized Function: " + str);

      return;

    default:
      throw new ExpressionParseException("Unrecognized Token Type");
    }
  }


  /**
    Create a Function token with symbol type specified as SYM_EXPRNODE,
    and an expression node.
    */
  public Symbol(int type, ExprNode en) {
    stype = type;
    enode = en;
  }


  /**
    Returns true if the symbol represents the end of an expression.
    */
  public boolean isEOE() {
    return (stype == SYM_TERMINAL) && (ttype == TK_EOE);
  }


  /**
    Returns true if the symbol represents an expression node.
    */
  public boolean isExprNode() {
    return (stype == SYM_EXPRNODE);
  }


  /**
    Returns true if current symbol has lower precedence than symbol a.
    */
  public boolean less(Symbol a) throws ExpressionParseException {
    if ((this.stype != SYM_TERMINAL) || (a.stype != SYM_TERMINAL))
      throw new ExpressionParseException(ExpressionParseException.EXPECT_TERMINAL);

    switch(precTab[precTabIndex(this.ttype)][precTabIndex(a.ttype)]) {
    case '<':
      return true;

    case '>':
    case '=':
      return false;

    default:
      throw new ExpressionParseException(ExpressionParseException.UNEXPECTED_SYMBOL);
    }
  }


  /**
    Returns true if current symbol has the same precedence as symbol a.
    */
  public boolean equal(Symbol a) throws ExpressionParseException {
    if ((this.stype != SYM_TERMINAL) || (a.stype != SYM_TERMINAL))
      throw new ExpressionParseException(ExpressionParseException.EXPECT_TERMINAL);

    switch(precTab[precTabIndex(this.ttype)][precTabIndex(a.ttype)]) {
    case '=':
      return true;

    case '>':
    case '<':
      return false;

    default:
      throw new ExpressionParseException(ExpressionParseException.UNEXPECTED_SYMBOL);
    }
  }


  /**
    Returns true if current symbol has higher precedence than symbol a.
    */
  public boolean greater(Symbol a) throws ExpressionParseException {
    if ((this.stype != SYM_TERMINAL) || (a.stype != SYM_TERMINAL))
      throw new ExpressionParseException(ExpressionParseException.EXPECT_TERMINAL);

    switch(precTab[precTabIndex(this.ttype)][precTabIndex(a.ttype)]) {
    case '>':
      return true;

    case '=':
    case '<':
      return false;

    default:
      throw new ExpressionParseException(ExpressionParseException.UNEXPECTED_SYMBOL);
    }   
  }


  /**
    Returns the string representation of the current symbol.
    */
  public String toString() {
    Double dval;

    switch (stype) {
    default:
    case SYM_UNKNOWN:
      return "Unknown symbol";

    case SYM_EXPRNODE:
      try {
	return ExprDouble.toString(enode.eval());
      } catch (Exception e) {
	return e.getMessage();
      }

    case SYM_TERMINAL:
      switch(ttype) {
      default:
	return "Unknown terminal";

      case TK_EOE:
	return "";

      case TK_NUMBER:
	return ExprDouble.toString(nval);

      case TK_VARIABLE:
	return variable.toString();

      case TK_BRACKETL:
	return "(";

      case TK_BRACKETR:
	return ")";

      case TK_PLUS:
      case TK_MINUS:
      case TK_MULT:
      case TK_DIV:
      case TK_POW:
      case TK_FUNCTION:
	return func.toString();
      }
    }
  }
}
